---
--    Copyright (c) 2014 by SAS Institute Inc., Cary, NC, USA.

--- This script/module contains Lua environment initialization logic for the stress and rrr product.
--  It is intended that all Lua code contained within the product should be executed 
--  after this script has run within the SAS session.
--  @module risk.stress.init
local M={} 


--========================================================================================
-- Set up access to our product's message files
--========================================================================================

--- The global get_message_file() function exists to provide a product-specific mapping
-- of message keys to message files, used by the risk.utils.sas_msg.get_message() function.
-- @param msg_key Message key
get_message_file = function(msg_key)
   if not msg_key then return end
   
   msg_key=msg_key:lower()
   if string.starts_with(msg_key, "sst_")  then
      return "SASHELP.stressluamsg"
   elseif string.starts_with(msg_key, "rrr_")  then
      return "SASHELP.rrrluamsg"
   elseif string.starts_with(msg_key, "utils_")  then
      return "SASHELP.rskluamsg"
   end
end


--========================================================================================
-- Set up Logging functions & levels
-- CAUTION: Turning on log level 4 and especially 5 will create a very large log - this is
-- in general a bad idea. You can use the set_log_level() function to set the log level to
-- dynamically (turn it up / down within the code) to avoid having the call return trace
-- on for the entire run.
--========================================================================================
local LEVELS = {NOTES = 1, SAS_SUBMIT = 2,DEBUG = 3, TRACE = 4, CALL_RETURN_TRACE=5 }
local LOG_LEVEL = sas.symgetn('LOG_LEVEL') or 0

local function get_time_str()
   return sas.putn(sas.time(),"time12.3")
end

function set_log_level(level)
   LOG_LEVEL = level
   if LOG_LEVEL < LEVELS.NOTES then sas.setoption("NOTES", false) else sas.setoption("NOTES", true) end
   if LOG_LEVEL < LEVELS.SAS_SUBMIT then sas.set_quiet(true) end 

   -- Enable Lua call/return tracing when appropriate
   if LOG_LEVEL >= LEVELS.CALL_RETURN_TRACE then
      -- print('setting trace hook')
      debug.sethook(trace_call_return_ftn, 'cr')
   end
   return
end

is_debug_enabled = function()
  return LOG_LEVEL >= LEVELS.DEBUG
end
is_trace_enabled = function()
  return LOG_LEVEL >= LEVELS.TRACE
end
Debug = function(...)
  if is_debug_enabled() then
     local arg = {...}
     -- find largest index as nils break ipairs
     local max = 0
     for i,v in pairs(arg) do
         if i > max then max = i end
     end
     local p = ''
     for i=1,max  do
        local str
        local v = arg[i]
        if type(v) == "table" then
           str = table.tostring(v) 
        else
           str = tostring(v)
        end
        p = p .. str .. " "
      end
      print(get_time_str() .. ' DEBUG: '.. p)
    end
end
Trace = function(...)
  if is_trace_enabled() then 
     local arg = {...}
     -- find largest index as nils break ipairs
     local max = 0
     for i,v in pairs(arg) do
         if i > max then max = i end
     end
     local p = ''
     for i=1,max  do
        local str
        local v = arg[i]
        if type(v) == "table" then
           str = table.tostring(v) 
        else
           str = tostring(v)
        end
        p = p .. str .. " "
      end
      print(get_time_str() .. ' TRACE: '.. p)
  end
end
Debug('LOG_LEVEL is ', LOG_LEVEL)


-- To keep from polluting the log with some functions that provide little value most of the time, we are
-- removing their trace points from going to the log.
local IGNORE_FILES = {['lunit.lua']=1}

local IGNORE_FUNCTIONS = {
    ['sasext.lua:tostring']=1, 
    ['sasext.lua:assert']=1,
    ['sasext.lua:do_token_replace']=1,
    ['global_init.lua:is_debug_enabled']=1,
    ['global_init.lua:is_trace_enabled']=1,
    ['global_init.lua:get_time_str']=1,
    ['global_init.lua:Debug']=1,
    ['global_init.lua:Trace']=1,
    ['sasext.lua:submit1']=1,
    ['sasext.lua:do_submit']=1,
    ['sasext.lua:preprocess_code']=1,
    ['sasext.lua:next_snippet']=1,
    ['sasext.lua:for iterator']=1
                             }

-- Returns nil if the calling ftn should not be logged, otherwise returns a string representing the name of the function in its file
local calling_ftn_name = function()
   local out=nil  -- default to tracepoint should not be logged
   local info= debug.getinfo(3,"nS")  -- 3 = get caller of caller of this function
   local path, ftn_file, ext
   ftn_file=""
   if info.name then 
      path, ftn_file, ext = string.match(info.short_src, "(.-)([^\\/]-%.?([^%.\\/]*))$")
   end

   -- only log non-C functions
   if not string.starts_with(ftn_file, '[C]') then
      local ftn_name = info.name or ""
      out = ftn_file .. ':' .. ftn_name
      if IGNORE_FILES[ftn_file] or IGNORE_FUNCTIONS[out] then
        out = nil
     end
   end 
   
   return out 
end

local get_args_str = function(skip_temps)
   local s = ''
   local i = 1
   while true do
      local name, value = debug.getlocal(3, i) -- 3 = get local vars at call time for caller of caller of get_args_str() ... i.e. caller of debug hook
      if not name or (skip_temps and name == '(*temporary)') then
         -- stop going through name value pairs, you are at the end 
         break 
      end
      
      -- comma separate the args.  do this even if skipping 1 below to give an indication to the user we skipped a type we don't print
      if #s > 0 then
         s = s .. ', '
      end
      
      -- Skip printing certain types of variables.  They provide little value (other than their address) and pollute the sas log quite a bit causing a lot of spilling onto next lines because the log is only a fixed length
      local value_type = type(value) 
      if value_type ~= 'table'  and value_type ~='function' then
         s = s .. name .. '=' .. tostring(value);
      end
      i = i + 1
    end
    return s
end

local trace_call_return_ftn = function(event)
   -- print ('in hooked trace ftn')
   if event == 'call' or event == 'tail call' then
      local ftn = calling_ftn_name()
      if ftn then
         local s = '    args: ' .. get_args_str(true)
         print(get_time_str() .. ' >> ' .. ftn .. s)
      end
   elseif event == 'return' then
      local ftn = calling_ftn_name()
      if ftn then
         local s = get_args_str()
         if #s> 0 then   -- only include if there are any vars at the return
            s = '    vars-at-return: ' .. s
         end
         print(get_time_str() .. ' << ' .. ftn .. s)
      end
   end
end

-- Enable Lua call/return tracing when appropriate
if LOG_LEVEL >= LEVELS.CALL_RETURN_TRACE then
   -- print('setting trace hook')
   debug.sethook(trace_call_return_ftn, 'cr')
end
set_log_level(LOG_LEVEL)



-- Commented out until we can fix sasstr
--- Set up to use the multi-byte string capabilities provided by sasstr -- this allows all of our RMX Lua string code to "just work"
--  in the encoding that the SAS session runs in -- which is what a user expects.
--  The following needs to occur for this to happen completely:
--  1)  All functions provide by the string library are being replaced (when used by RSK/RMX) by multi-byte equivalents.
--      This then causes string.len(mystr) to call into sasstr instead of the single-byte Lua string library.
--   ... the following must be done in our Lua C code ... can't be done from here ...
--  2)  All string functions need to be overridden using the metatable for the string type to point to sasstr. so that s:upper() goes to sasstr 
--      This then causes  mystr:len() to call into sasstr instead of the single-byte Lua string library
--  3)  All string operators need to be directed to functions in sasstr
--      This then causes  (str_a .. str_b) to perform the concatenation operation from sasstr instead of the single-byte Lua string library
--      
--  Commented out until sasstr is better tested and working as expected  
--local sasstr = require 'sasstr'
--sasstr.take_over()  -- there is no undoing this for a Lua VM instance once it's done

return M

