---
--    Copyright (c) 2014 by SAS Institute Inc., Cary, NC, USA.
-- 
--- metadata related routines - get and set properties
-- @module metadata
-- 
local M={}
require 'risk.common.global_init'

--- Searches and gets all properties for a given software component from metadata. 
--- Within sofwtare component it searches for -> PropertySets association -> Application COnfiguration -> SetProperties (this tree holds properties). 
-- @param software_comp_name - Name of software component as you see in SMC for. 
-- @return metadata_properties - A lua table with the properties for the component
-- @return uri_PropertySets - The uri for the property set being returned. Used to modify / add properties
-- @usage
-- local props, prop_set_uri = metadata.get_properties_from_soft_comp("Risk Finance Wkbnch Svr Cfg")
function M.get_properties_from_soft_comp(software_comp_name)
   local 
      prop_ds,
      i,
      nrows,
      data_exists,
      dsid,
      metadata_properties
   sas.submit([[
   data all_prop;
      length uri_software_component $256;
      uri_software_component='';
      software_component_found = metadata_getnobj("omsobj:SoftwareComponent?@Name contains '@software_comp_name@'",
                                                  1, uri_software_component);
      if software_component_found = 1 then do;
         PropertySets_found = 0;    /*ApplicationActions Tree found*/
         property_rc=1;
         property_count=1;
         PropertySets_rc = 1;
         PropertySets_count = 1;
         length 
            uri_PropertySets $256
            uri_SetProperties $256
            uri_AppCOnfigName $256.
            property_Name $256.
            uri_property_name $256.
            defaultvalue $256.
            ;
          appConfig_Rc=1;
          uri_AppCOnfigName='';
          uri_PropertySets = '';
    
         PropertySets_rc = metadata_getnasn(uri_software_component, "PropertySets", PropertySets_count, uri_PropertySets);
         call symput("uri_PropertySets",uri_PropertySets);
         /* get individual property */
         do while (property_rc>0) ;
            property_rc = metadata_getnasn(uri_PropertySets, "SetProperties", property_count,uri_property_name );
            property_count=property_count+1;
            property_name_rc=metadata_getattr(uri_property_name, "Name", property_Name);
            property_val_rc=metadata_getattr(uri_property_name, "DefaultValue", defaultvalue );
            if property_rc > 0 then output;
         end;
      end;
   run;
   ]], {software_comp_name = software_comp_name })
   local uri_PropertySets = sas.symget("uri_PropertySets")
   metadata_properties = {}
   data_exists = sas.exists("work.all_prop")
   if data_exists then
      prop_ds = sas.load_ds("work.all_prop")
      dsid = sas.open("work.all_prop")
      nrows = sas.nobs(dsid)
      sas.close(dsid)
   else
--       sas.print("%1z Unable to get metadata properties for %s",software_comp_name)
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_get_properties_from_soft_comp_1', 'NOQUOTE'),software_comp_name)
      return nil, nil
   end
   -- There are some datsset specific information entries in prop_ds that we don't want
   for i = 1,nrows do
     metadata_properties[i] = prop_ds[i]
   end
   return metadata_properties, uri_PropertySets
end

--- Assign a value to a software component property attribute (typically 'DefaultValue').
-- @param software_comp_name - The name of the software component
-- @param property_name - The name of the property whose attribute will be set
-- @param attribute_name - The name of the attribute (e.g. DefaultValue)
-- @param attribute_value - The value to assign to the attribute
--@return pass - Boolean whether or not the attribute was set
function M.soft_comp_property_setattr(software_comp_name, property_name, attribute_name, attribute_value)
   local rfw_properties = M.get_properties_from_soft_comp(software_comp_name)
   for i,prop in ipairs(rfw_properties) do
      if prop.property_name == property_name then
         local prop_rc = sas.metadata_setattr(prop.uri_property_name, attribute_name, tostring(attribute_value))
         if prop_rc >= 0 then
            return true
         else
            --print("%1z Unabble to set the attribute - metadata_setattr returned " .. prop_rc)
            return false
         end
      end
   end
   --print('%1z A property with the name ' .. (property_name or "<nil>") .. ' was not found.')
   return false
end

--- Add a new property to a software component. This is the same as going into the plugins tab in the smc,
--- selecting the software component, then going to properties -> advanced and adding one there.
-- @param software_comp_name - The name of the software component 
-- @param property_name - The name of the property to add
-- @param repository - The repository, defaults to Foundationget_properties_from_soft_comp
-- @return pass - boolean, whether or not the new property was created
function M.soft_comp_property_add(software_comp_name, property_name, repository)
   repository = repository or 'Foundation'
   local rfw_properties, uri_property_sets = M.get_properties_from_soft_comp(software_comp_name)
   for i,prop in ipairs(rfw_properties) do
      if prop.property_name == property_name then
         --print("%3z The property ".. property_name .. " already exts.")
         return true
      end
   end
   local property_name_uri
   property_name_uri = ''
   local prop_rc = sas.metadata_newobj('Property', property_name_uri, property_name, repository, uri_property_sets,"SetProperties" )
   if prop_rc < 0 then
      --print("%1z Unabble to create the property - metadata_newobj returned " .. prop_rc)
      return false
   end
   local prop_name_rc = sas.metadata_setattr(property_name_uri, 'PropertyName', property_name);
   if prop_name_rc < 0 then
      --print("%1z Unabble to set the name for the property - metadata_setattr returned " .. prop_name_rc)
      return false
   end
   local prop_usage_ver_rc = sas.metadata_setattr(property_name_uri, 'UsageVersion', tostring(0))
   if prop_usage_ver_rc < 0 then
      -- print("%1z Unabble to the version for the property - metadata_setattr returned " .. prop_usage_ver_rc)
      return false
   end
    --print("%3z Added property " .. property_name)
    return true
end

   
---  Get the end point rest URL and other details
-- @ param AppRestCompName - The rest component name (e.g. SASRiskandFinanceWorkbench REST)
-- @return hostname, port, protocol, appName, URL - the rest component details
function  M.get_rest_end_point_dtls(AppRestCompName)
local hostname, port, protocol
sas.submit([[
   %svcreg(type=REST, name=@AppRestCompName@);
   data svc;
      attrib RESTURL format=$256. informat=$256.;  
      str= symget("service_endpoint");
 
      parse=prxparse("/https?\:\/\/(.*)\:/");
      app_context=prxparse("/(https?)\:\/\/(.*)\:(\d+)\/(.*?)\//");
      if prxmatch(app_context, str) then do;  
         protocol = prxposn(app_context, 1, str);
         hostname = prxposn(app_context, 2, str);
         port = prxposn(app_context, 3, str);
         appName = prxposn(app_context, 4, str);
         CALL SYMPUTX('hostname', hostname ,'G');
         CALL SYMPUTX('port', port ,'G');
         CALL SYMPUTX('protocol', protocol ,'G');
         CALL SYMPUTX('appName', appName ,'G');
         
         parse1=prxparse("/(.*rest\/)/");
         url1_match=prxmatch(parse1, str);
         URL = prxposn(parse1, 1, str);
         CALL SYMPUTX('URL', URL ,'G');
      end;
   run;
   ]])
   local hostname = sas.symget("hostname")
   local port = sas.symget("port")
   local protocol = sas.symget("protocol")
   local appName = sas.symget("appName")
   local URL = sas.symget("URL");
   return hostname, port, protocol, appName, URL
end
return M 
