---
--    Copyright (c) 2016 by SAS Institute Inc., Cary, NC, USA.
-- 
--- REST service manipulation routines
--  you first need to set the url using set_url(). 
--  Set the authentication method with set_auth()
--  There is a table called M.htttp_codes that you can use to look up the return codes
--  returned from request()
--  @usage 
--[[ EXAMPLE USING AUTHENTICATION ON THE MIDTIER:
   %let log_level = 1;
   filename LUAPATH ( '!SASROOT/misc/stresssvr/lua/sas' '!SASROOT/stresssvr/sasmisc/lua/sas' );

   proc lua restart; 
      submit;
         local rest = require 'risk.common.rest_service'
         local JSON = require 'risk.common.json'
         local pass,json_out, code
         JSON.strictTypes = true

         print("rest.url: "..tostring(rest.url))
         pass = rest.set_url('http://myMachine.com:7980/SASRiskandFinanceWorkbench/rest/')
         print("Return from  set_url: "..tostring(pass))
         pass = rest.set_auth('token')
         pass, json_out, code = rest.request("GET", "projects")

         print("pass: "..tostring(pass))
         print("code: "..tostring(code).." "..rest.http_codes[code])
         print("json_out: "..table.tostring(JSON:decode(json_out)))

      endsubmit;
   run;
]]--
--[[ EXAMPLE USING TICKET GRANTING TICKET:
   %let log_level = 1;
   filename LUAPATH ( '!SASROOT/misc/stresssvr/lua/sas' '!SASROOT/stresssvr/sasmisc/lua/sas' );

   proc lua restart; 
      submit;
         local rest = require 'risk.common.rest_service'
         local JSON = require 'risk.common.json'
         local pass,json_out, code
         JSON.strictTypes = true

         print("rest.url: "..tostring(rest.url))
         pass = rest.set_url('http://myMachine.com:7980/SASRiskandFinanceWorkbench/rest/')
         print("Return from  set_url: "..tostring(pass))
         
         pass = rest.set_auth('ticket','sasdemo','Go4thsas')
         print("Return from set_auth: "..tostring(pass))
         
         pass = rest.set_tgt()
         print("Return from get-rest_tgt: "..tostring(pass))
         
         pass, json_out, code = rest.request("GET", "projects?")

         print("pass: "..tostring(pass))
         print("code: "..tostring(code).." "..rest.http_codes[code])
         print("json_out: "..table.tostring(JSON:decode(json_out)))        

      endsubmit;
   run; 
]]--

-- @module rest_service
local M={}
require 'risk.common.global_init'
local utils=require 'risk.common.rest_utils'
local string_utils=require 'risk.common.string_utils'
local json=require 'risk.common.json'

-- This is the base URL for all calls. 
-- To be set with a call to set_url
M.url=nil
M.tgt_url = nil

-- The authentication method can be 'ticket' or 'token', 'basic', or 'none'
-- To be set with a call to set_auth. 
M.auth = nil
M.service_ticket=nil
M.tgt=nil
local username = nil
local password = nil

-- Whether or not to use cookies. Default is true which saves from making repeated calls for a service ticket.
M.use_cookies = true

-- Controls length of printed strings when Debugging. 
M.max_print = 32000
-- Assign filerefs  to be used later so each individual function doesn't need to do this. 
local file_sep = utils.get_file_sep();  
local temp_folder = sas.getoption('work')   
sas.submit([[
   filename _hin_  "@temp_folder@@file_sep@_hin.txt";
   filename hout "@temp_folder@@file_sep@_hout.txt";
   filename _in_   "@temp_folder@@file_sep@_in.txt";
   filename out  "@temp_folder@@file_sep@_out.txt";
]], {temp_folder = temp_folder, file_sep = file_sep })

--- Set the URL and filenames for http requests.
-- @param in_url  - a string containing the url
function M.set_url(in_url)   
   -- Assign the url to M.url
   M.url = in_url
   M.tgt_url = string.match(M.url, "^(http[s]?://.+/SAS).*$") .. 'Logon/v1/tickets/'
   M.service_ticket = nil
   Debug("URL set to " , M.url)
   return true
end

---Sets the authentication method. Valid options are token, ticket, and basic
--- If you use ticket generating ticket, you need to call set_tgt before making calls to the api
-- @param auth_method - the method to use.
-- @param user_name[OPT] - username for basic auth and getting a tgt
-- @param pass_word[OPT] - password for basic auth and getting a tgt
-- @return pass - boolean whether or not the value was set.
-- @usage
-- local rest = require 'risk.common.rest_service'
-- local pass = rest.set_auth('ticket')
function M.set_auth(auth_method,user_name,pass_word)
   local method = sas.lowcase(auth_method)
   M.auth = nil
   username = nil
   password = nil
   if method == "token" then
      M.auth = "token"
   elseif method == "ticket" then
      M.auth = "ticket"
      if user_name and user_name:len() ~= 0 then username = user_name end
      if pass_word and pass_word:len() ~= 0 then password = pass_word end
   elseif method == "basic" then
      if not pass_word or pass_word:len() == 0 then
         sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_no_pass_for_basic', 'NOQUOTE'))
         return false
      elseif not user_name or user_name:len() == 0 then
         sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_no_username_for_basic', 'NOQUOTE'))
         return false
      end
      M.auth = "basic"
      username = user_name
      password = pass_word
   elseif method == 'none' then
      M.auth = 'none'
   else
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_not_valid_auth', 'NOQUOTE'),auth_method)
      if M.auth then
         sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_using_prev_auth', 'NOQUOTE'),M.auth)
      end
      return false
   end
   return true
end

--- Set a ticket granting ticket. Pass in nil to use get_tgt_cas( )
--- to generate the ticket.
-- @param tgt[OPT] - the value to set tgt to. If nil, get_tgt_cas is called.
-- @return pass - boolean, whether or not the call was succesful
function M.set_tgt(tgt)
   Debug("Calling: set_tgt")
   if M.auth ~= 'ticket' then 
      Debug('Calling set_tgt but auth is not ticket. M.auth is: '..tostring(M.auth))
      return false 
   end
   if not M.url or M.url:len() == 0 then
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'),M.auth)
      return false
   end
   if not tgt or tgt:len() == 0 then
      local pass = M.get_tgt_cas( username, password )
      if not pass then return false end
   else
      M.tgt = tgt
   end
   return true
end

--- Get a CAS REST service ticket. This returns 'true' if successful and sets the value
--- of 'rest_ticket_granting_ticket' to be used by get_service_ticket.
-- @param username  - a string containing the user name
-- @param password  - a string containing the user password
function M.get_tgt_cas( username, password )
   Debug("Calling: get_tgt_cas")
   if M.auth ~= 'ticket' then return false end
   if not M.url or M.url:len() == 0 then
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'),M.auth)
      return false
   end
    utils.write_fileref("_hin_", 'accept:text/plain')
    utils.write_fileref_serialized("_in_", 'username='..utils.urlencode(username)..'&password='..utils.urlencode(password))
    
    -- Get a ticket granting ticket - TGT 
    sas.submit([[
       proc http 
          url=%nrstr("@tgt_url@")
          headerin=_hin_ 
          headerout=hout
          in=_in_ 
          out=out
          method='post' 
          ct='text/plain';
       run;
     ]],{tgt_url = M.tgt_url})
     
     local ret = utils.read_fileref('hout')
     
     -- find the location keyword
     local indx = sas.index(ret, 'Location: ')
     if indx  < 1 then return false end
     ret = sas.substr(ret, indx+10)
     indx = sas.index(ret, '\n')
     ret = sas.trim( sas.substr(ret, 1, indx-1))
     indx = string.find(ret,"SASLogon/v1/tickets/")
     M.tgt = string.sub(ret,indx+20)    
     
     Debug("TICKET GRANTING TICKET REQUEST:")
     Debug("URL: ", M.tgt_url)
     Debug("Header In:", utils.read_fileref('_hin_'))
     Debug("Header Out:" , utils.read_fileref('hout'))
     Debug("Body Out:" , utils.read_fileref('out'))
     Debug("Ticket Granting Ticket:" , M.tgt )
     if sas.index(M.tgt, 'TGT-') == 0 then 
        sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_failed_to_get_tgt', 'NOQUOTE'))
        M.tgt = nil
        return false
     end
    return true
end
--- GET operation on a REST service. Must have a ticket granting ticket first. 
--- This is called by request_json_tgt() for you so you shouldn't have to call this.
-- @param request - the service url for which the service ticket is requested
-- @return pass -  boolean whether the header said OK
function M.get_service_ticket( request )
   Debug("Calling: get_service_ticket")
   if not M.url or M.url:len() == 0 then
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'),M.auth)
      return false
   end
   if not request or request:len() == 0 then
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_request_empty', 'NOQUOTE'))
      return false
   end
   if not (M.tgt or M.set_tgt()) then 
     sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_need_tgt', 'NOQUOTE'))
     return false, nil, nil
   end

   local service_url = 'service='..utils.urlencode(M.url..request)
   if sas.index(service_url, '?') == 0 then
      service_url = service_url..'?'
   end
   utils.write_fileref("_in_", service_url )
   sas.submit( [[
     proc http 
        url=%nrstr("@tgt_url@@tgt@") 
        headerout=hout
        in=_in_
        out=out
        method='post' 
        ct='text/plain';
      run;
   ]],{tgt_url=M.tgt_url,tgt=M.tgt})
   M.service_ticket = utils.read_fileref('out')
     
   Debug("SERVICE TICKET REQUEST:")
   Debug("Header Out:" , utils.read_fileref('hout'))
   Debug("Body In:" , utils.read_fileref('_in_'))
   Debug("Body Out:" , utils.read_fileref('out'))
   Debug("Ticket:",M.service_ticket )
  
   if sas.index(M.service_ticket, 'ST-') == 0 then 
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_get_service_ticket_1', 'NOQUOTE'))
      M.service_ticket = nil
      return false
   end 
    return true;
 end
 
--- Submit a request to M.url using the M.auth authentication method
--- Output is captured in the filerefs out and hout
-- @param action - string, GET, POST, PUT, etc.
-- @param request - the portion of the URL that follows M.url
-- @param body[OPT] - text for the body. If nil, then no body is passed
-- @param header_in[OPT] - text for the input header. If nil, then no header is passed (proc http default is used)
-- @param content_type[OPT] - This is the content type. You can pass this in with the header but proc http has a separate argument for this
-- @param out[OPT] - The fileref to write the output to - defaults to 'out' - If not specified the body will be returned from the call as a string.
-- @param hout[OPT] - The fileref to write the returned header to. Defaults to 'hout'
-- @return pass - whether or not the HTTP return code was in the 200's (200 OK, 201 CREATED, etc)
-- @return json - The output from the call as text if no fileref is specified
-- @return code - the actual http return code
-- @usage
-- pass, json = rest.request( "get", "config/hierarchies")
-- local pass, json = rest.request( "get", "ruleFlows/".. ruleFlowId.."/deployments", nil , 
--                                  "Accept:application/vnd.sas.businessrules.deployments+json;version=1")
function M.request( action, request, body, header_in, content_type, out, hout )
   if not M.url or M.url:len() == 0 then
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'),M.auth)
      return false,nil,nil
   end
   if not action or not request then
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_need_request_and_action', 'NOQUOTE'))
      return false, nil, nil
   end
   local http_options = ""
   local url_str
   if M.auth == "token" then 
      http_options = "http_tokenauth\nclear_cache"
      url_str = M.url..request
   elseif M.auth == "ticket" then
      local ticket_format
      if sas.index(request, '?') ~= 0 then
         ticket_format = '&'
      else
         ticket_format = '?'
      end      
      if not M.service_ticket or not M.use_cookies then
         if not M.get_service_ticket( request ) then 
            sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_could_not_get_ticket', 'NOQUOTE'))
            return false, nil, nil 
         end        
      url_str = M.url..request..ticket_format.."ticket="..M.service_ticket
      http_options = "clear_cache"
      else  -- Not First use cookies    
         url_str = M.url..request..ticket_format.."ticket="..M.service_ticket
         http_options = ""
      end
      
   elseif M.auth == 'basic' then
      url_str = M.url..request
      local username = string_utils.escape_sas_string_value(username)
      local password = string_utils.escape_sas_string_value(password)
      http_options = 'AUTH_BASIC webusername="'..tostring(username)..'" webpassword="'..tostring(password)..'"\nclear_cache'
   elseif  M.auth == 'none' then
      url_str = M.url..request
      http_options = "clear_cache"
   else
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_not_valid_auth', 'NOQUOTE'),auth_method)
      return false, nil, nil
   end
   
   local body_arg, header_arg, ct_arg
   
   -- Set the 'in' argument to proc http
   if not body then 
      body_arg = " " 
   else
      utils.write_fileref("_in_",body)
      body_arg = "in = _in_"
   end
   -- Set the 'headerin' argument to proc http
   if not header_in then 
      header_arg = " " 
   else
      utils.write_fileref("_hin_",header_in)
      header_arg = "headerin = _hin_"
   end

   -- Set the content type and out arguments
   if not content_type then ct_arg = " " else ct_arg = "ct = '"..content_type.."'" end
   if not out then out = "out" end
   if not hout then hout = "hout" end
   utils.write_fileref(out,"_NOT_SET_BY_PROC_HTTP_")
   utils.write_fileref(hout,"_NOT_SET_BY_PROC_HTTP_")
   
  -- If in debug, print the body and header
   if header_in and is_debug_enabled() then
      Debug('HEADER IN')
      print(string.sub(header_in,1,M.max_print))
   end 
   if body and is_debug_enabled() then
       Debug('BODY IN')
      print(string.sub(body,1,M.max_print))
   end  

   sas.submit([[
      proc http
          @http_options@      
          url=%nrstr("@url_str@") 
          @header_arg@  
          headerout=@hout@
          @body_arg@ 
          out=@out@
          @ct_arg@
          method='@action@';
      run;
   ]])
   Debug('HEADER OUT')
   Debug(utils.read_fileref(hout))
   Debug('BODY OUT')
   Debug(utils.read_fileref(out))
   
   
   --Grab the http return code
   local header_string = utils.read_fileref(hout)
   local header_line_one = sas.scan(header_string,1,'\n\r')
   local code = tonumber(sas.scan(header_line_one,2,' '))
   local pass  = (code >= 200 and code < 300)
   local return_body = utils.read_fileref(out)

   return pass, return_body, code
end


-- @Section HHTP Return Codes Table
M.http_codes={}
M.http_codes[100]='Continue'
M.http_codes[101]='Switching Protocols'
M.http_codes[102]='Processing (WebDAV; RFC 2518)'
M.http_codes[200]='OK'
M.http_codes[201]='Created'
M.http_codes[202]='Accepted'
M.http_codes[203]='Non-Authoritative Information (since HTTP/1.1)'
M.http_codes[204]='No Content'
M.http_codes[205]='Reset Content'
M.http_codes[206]='Partial Content (RFC 7233)'
M.http_codes[207]='Multi-Status (WebDAV; RFC 4918)'
M.http_codes[208]='Already Reported (WebDAV; RFC 5842)'
M.http_codes[226]='IM Used (RFC 3229)'
M.http_codes[300]='Multiple Choices'
M.http_codes[301]='Moved Permanently'
M.http_codes[302]='Found'
M.http_codes[303]='See Other (since HTTP/1.1)'
M.http_codes[304]='Not Modified (RFC 7232)'
M.http_codes[305]='Use Proxy (since HTTP/1.1)'
M.http_codes[306]='Switch Proxy'
M.http_codes[307]='Temporary Redirect (since HTTP/1.1)'
M.http_codes[308]='Permanent Redirect (RFC 7538)'
M.http_codes[400]='Bad Request'
M.http_codes[401]='Unauthorized (RFC 7235)'
M.http_codes[402]='Payment Required'
M.http_codes[403]='Forbidden'
M.http_codes[404]='Not Found'
M.http_codes[405]='Method Not Allowed'
M.http_codes[406]='Not Acceptable'
M.http_codes[407]='Proxy Authentication Required (RFC 7235)'
M.http_codes[408]='Request Timeout'
M.http_codes[409]='Conflict'
M.http_codes[410]='Gone'
M.http_codes[411]='Length Required'
M.http_codes[412]='Precondition Failed (RFC 7232)'
M.http_codes[413]='Payload Too Large (RFC 7231)'
M.http_codes[414]='URI Too Long (RFC 7231)'
M.http_codes[415]='Unsupported Media Type'
M.http_codes[416]='Range Not Satisfiable (RFC 7233)'
M.http_codes[417]='Expectation Failed'
M.http_codes[418]="I'm a teapot (RFC 2324)"
M.http_codes[421]='Misdirected Request (RFC 7540)'
M.http_codes[422]='Unprocessable Entity (WebDAV; RFC 4918)'
M.http_codes[423]='Locked (WebDAV; RFC 4918)'
M.http_codes[424]='Failed Dependency (WebDAV; RFC 4918)'
M.http_codes[426]='Upgrade Required'
M.http_codes[428]='Precondition Required (RFC 6585)'
M.http_codes[429]='Too Many Requests (RFC 6585)'
M.http_codes[431]='Request Header Fields Too Large (RFC 6585)'
M.http_codes[451]='Unavailable For Legal Reasons'
M.http_codes[500]='Internal Server Error'
M.http_codes[501]='Not Implemented'
M.http_codes[502]='Bad Gateway'
M.http_codes[503]='Service Unavailable'
M.http_codes[504]='Gateway Timeout'
M.http_codes[505]='HTTP Version Not Supported'
M.http_codes[506]='Variant Also Negotiates (RFC 2295)'
M.http_codes[507]='Insufficient Storage (WebDAV; RFC 4918)'
M.http_codes[508]='Loop Detected (WebDAV; RFC 5842)'
M.http_codes[510]='Not Extended (RFC 2774)'
M.http_codes[511]='Network Authentication Required (RFC 6585)'
M.http_codes[103]='Checkpoint'
M.http_codes[420]='Method Failure (Spring Framework)'
M.http_codes[450]='Blocked by Windows Parental Controls (Microsoft)'
M.http_codes[499]='Token Required or Request has been forbidden by antivirus]'
M.http_codes[509]='Bandwidth Limit Exceeded (Apache Web Server/cPanel)'
M.http_codes[530]='Site is frozen'
M.http_codes[440]='Login Timeout'
M.http_codes[449]='Retry With'
M.http_codes[451]='Redirect'
M.http_codes[444]='No Response'
M.http_codes[495]='SSL Certificate Error'
M.http_codes[496]='SSL Certificate Required'
M.http_codes[497]='HTTP Request Sent to HTTPS Port'
M.http_codes[499]='Client Closed Request'
M.http_codes[520]='Unknown Error'
M.http_codes[521]='Web Server Is Down'
M.http_codes[522]='Connection Timed Out'
M.http_codes[523]='Origin Is Unreachable'
M.http_codes[524]='A Timeout Occurred'
M.http_codes[525]='SSL Handshake Failed'
M.http_codes[526]='Invalid SSL Certificate'

return M
