---
--    Copyright (c) 2016 by SAS Institute Inc., Cary, NC, USA.
-- 

--- Utility functions for REST module 
-- @module rest_service
local M={}

require 'risk.common.global_init'

--- Get the file separator for the operating system 
-- @return file_sep - either '\' or '/'
function M.get_file_sep()
  local host = sas.symget("sysscp")
  if sas.index("WIN", host) ~= 0 then
    file_sep = "\\"
  else
    file_sep = "/"
  end
  return file_sep
end
  
--- Read a file referenced by fileref into a string.
-- @param fileref - either a string or a fileref from sasxx.assign()
function M.read_fileref( fileref )
   if type(fileref) == "string" then
       fileref =  sasxx.new(fileref)
   end
   local path = fileref:info().path   
   if not path then
      fileref:deassign()
      return false, "Couldn't open file referenced by "..tostring(fileref).." for read."   
   end

   local BUFSIZE = 2^13
   local f = io.open(path,"rb")
   if not f then
      return false, "Couldn't open file referenced by "..tostring(fileref).." for read."   
   end   

   local contents = ""
   while true do
      local bufread = f:read(BUFSIZE)
      if not bufread then break end
	  contents = contents..bufread
   end
   f:close()
   return contents
end


--- Read a file referenced by a filename into a string.
-- @param filename - fully qualified pathname to the file
function M.read_file( filename )
   local fileref = sasxx.assign( filename)
   if not  fileref then return nil end
   -- print(table.tostring(fileref:info()))
  
  local contents = M.read_fileref(fileref)
  fileref:deassign()

  return contents
end

--- Write a file referenced by fileref from a string with carriage returns
-- @param fileref - either a string or a fileref from sasxx.assign()
-- @param txt  - the string being written to the file
function M.write_fileref( fileref, txt )
    if type(fileref) == "string" then
       fileref =  sasxx.new(fileref)
   end   
   local path = fileref:info().path   
   if not path then
      fileref:deassign()
      return false, "Couldn't open file referenced by "..tostring(fileref).." for write."   
   end   
   local f = io.open(path,"wb")
   if not f then       
      return false, "Couldn't open file: "..path.." for write."
   end
   f:write(txt)
   f:close()
   return true
end

--- Write a file referenced by fileref from a string serialized
-- @param fileref - either a string or a fileref from sasxx.assign()
-- @param txt  - the string being written to the file
function M.write_fileref_serialized( fileref, txt )
    if type(fileref) == "string" then
       fileref =  sasxx.new(fileref)
   end

   local file_handle = fileref:open("w-o","f", 1)
   if not  file_handle then
      fileref:deassign()
      return false, "Couldn't open file for write." 
   end

   local lines = string.split(txt, "\n")
   for i, line in ipairs(lines) do
       for i = 1, string.len(line) do
           file_handle:write(string.sub(line, i, i) )
       end
  end

  file_handle:close()
  return true
end

--- Write a file referenced by a filename from a string with carriage returns
-- @param fileref - either a string or a fileref from sasxx.assign()
-- @param txt  - the string being written to the file
function M.write_file( filename, txt )
   local fileref = sasxx.assign( filename)
   if not  fileref then return false, "Couldn't open file location."  end

   local pass, msg = M.write_fileref( fileref, txt )
   fileref:deassign()

   return pass, msg
end

--- Append 2 files byte by byte, file2 is added to the end of file1 (file1 is modified)
-- @param file1 - The first file (fileref)
-- @param file2 - The second file (fileref)
function M.append_file_by_bytes( file1, file2 )

   sas.submit([[
    /* copy the file byte-for-byte */ 
    data _null_;
       length 
              fileID1 8 
              fileID2 8
              ;
       fileID1 = fopen("@file1@",'A',1,'B');
       fileID2 = fopen("@file2@",'I',1, 'B');
       rec = '20'x;
        
       do while(fread(fileID2)=0);
          rc = fget(fileID2,rec,1);
          rc = fput(fileID1, rec);
          rc = fwrite(fileID1);
       end;
        
       rc = fclose(fileID1);
       rc = fclose(fileID2);
    run;
    
   ]], {file1 = file1, 
        file2 = file2})    
end
--- Encode a  URL string appropriate for asking for a service ticket
-- @param str String containing a URL to encode
function M.urlencode( str )
    if (str) then
        str = string.gsub (str, "\n", "\r\n")
        str = string.gsub (str, "([^%w ])",
            function (c) return string.format ("%%%02X", string.byte(c)) end)
        str = string.gsub (str, " ", "+")
    end
    return str    
end

---Used for searching inside json or table (including json with nested arrays. Total 1 or max 3 attribute you can search. We perform AND on all the search. 
--Very useful in searching through complex json. It returns the table where it find matching attributeName{n} and attributeName{v}  
--  @param jsonStringorTable : Json string or table
--  @param attribute : what key to search in Json or table : 
--  @param attributeName1 : Further in attribute json which key to search 
--  @param attributeValue1 : Further in attribute json which value for the above key to search
--  @param attributeName2 : Further in attribute json which key to search 
--  @param attributeValue2 : Further in attribute json which value for the above key to search
--  @param attributeName3 : Further in attribute json which key to search 
--  @param attributeValue3 : Further in attribute json which value for the above key to search
--  Example: Lets say we want to search in a json array which further has nested json array. I 
----local projectFilters=rfw.project_get_filter_details("Taxonomy test")
--print(Json:encode(projectFilters))
--json_or_table_filter(projectFilters,'items.lineItems[]',"memberName","004 - Bulgaria")
--Note that if line item i further array please specify []
--
--
--   
function M.json_or_table_filter(jsonStringorTable,attribute,attributeName1,attributeValue1,attributeName2,attributeValue2,attributeName3,attributeValue3)
   local jsonTable , response
   if type(jsonStringorTable) ~= "table" then 
      --assuming it is json string we need to decode to table 
      jsonTable=Json:decode(jsonStringorTable)
   else 
      jsonTable=jsonStringorTable
   end

   local dot='.'
   local array = {}
   local arrayCopy={}
   if attribute == nil or attribute == "" then 
      dot=''
      array = {""}
   else
      array=string.split(attribute , '.')
   end
   arrayCopy=array

   sas.submit([[
      filename file TEMP;
      %let filePath = %sysfunc(pathname(file));
   ]])

   local tmpFile=sas.symget("filePath")
   Debug("tmpFile:: "..tmpFile)
   local fileref =  sasxx.new("file")
   local f = fileref:open("w")

   --f = assert (io.tmpfile ()) -- open temporary file
   --f = assert (io.open (tmpFile,'w')) -- open temporary file
   for i , tmp in pairs(array) do 
      if ( i == 1) then 
         f:write("function foo(jsonTable) \n")
         f:write("local tab={} \n")
         f:write("for i, v  in pairs(jsonTable"..dot..tmp..") do \n")
      else 
         if string.match(tmp,"%[%]$") then 
            tmp=string.gsub(tmp,"%[%]","")
            f:write("for i, v  in pairs(v."..tmp..") do  \n ")
         else 
            f:write("v = v."..tmp.." \n" )
            table.remove( arrayCopy, i )
         end 
      end
   end

   for i , tmp in pairs(arrayCopy) do 
      if ( i == 1) then 
         --f:write("print(v) \n")
         f:write("table.insert(tab, v) \n")
      end
      f:write("end \n")
   end
   --function foo ends
   f:write("return tab \n end \n")
   f:close()

   local fileContent =M.read_file(tmpFile)
   Debug("##File content::### \n"..fileContent.."## \n")  -- print out

   local x =assert(loadfile(tmpFile))
   x()
   local resp=foo(jsonTable)
   --print(Json:encode(resp))
   local found=0

   if attributeName1 == nil and attributeName2 == nil and  attributeName3 == nil then 
      return resp
   end

   for i ,v in pairs(resp) do
      if attributeName1 ~= nil and attributeName2 == nil and  attributeName3 == nil then 
         if string.upper(v[attributeName1]) ==  string.upper(attributeValue1) then 
            response=v
            found=1
            break
         end  
      end 

      if attributeName1 ~= nil and attributeName2 ~= nil and  attributeName3 == nil then 
         if string.upper(v[attributeName1]) ==  string.upper(attributeValue1) and string.upper(v[attributeName2]) ==  string.upper(attributeValue2) then 
            response=v
            found=1
            break
         end
      end   

      if attributeName1 ~= nil and attributeName2 ~= nil and  attributeName3 ~= nil then 
         if string.upper(v[attributeName1]) ==  string.upper(attributeValue1) and string.upper(v[attributeName2]) ==  string.upper(attributeValue2) and string.upper(v[attributeName3]) ==string.upper(attributeValue3) then 
            response=v
            found=1
            break
         end
      end
   end

   if found == 1 then 
     return response
   else 
     sas.print("WARN: Cannot find one or more of the attributes or value in response.")
     return nil
   end
end 

return M
