---
--    Copyright (c) 2016 by SAS Institute Inc., Cary, NC, USA.
-- 
--- REST service manipulation routines - these are wrappers for the routines in the rest_service module in 
--  risk/common. The purpose of these modifications is to allow previous code written using the original
--  rest_service module in rfw to still function. 

-- @module rest_service
local M={}
require 'risk.common.global_init'
local rest_common=require 'risk.common.rest_service'
local utils=require 'risk.common.rest_utils'
local string_utils=require 'risk.common.string_utils'
local json=require 'risk.common.json'

-- Synch the values of the objects in rest_common with those in M so they are available to rfw_api when it uses this module.
local function synch_rest_settings()
   M.url = rest_common.url
   M.tgt_url = rest_common.tgt_url
   M.auth = rest_common.auth
   M.service_ticket = rest_common.service_ticket
   M.tgt = rest_common.tgt
   M.use_cookies = rest_common.use_cookies
end
synch_rest_settings()

-- Default auth is ticket when we are running in an rfw model group or script, otherwise it is token.
local sst_tgt = sas.symget("sst_tgt")
if sst_tgt and sst_tgt:len()>0 then
   rest_common.auth="ticket"
   rest_common.tgt=sst_tgt
else
    rest_common.auth="token"
end

-- Previous version used filerefs 'in' and 'hin' so make these available
local file_sep = utils.get_file_sep();  
local temp_folder = sas.getoption('work')   
sas.submit([[
   filename hin  "@temp_folder@@file_sep@hin.txt";
   filename in   "@temp_folder@@file_sep@in.txt";
]], {temp_folder = temp_folder, file_sep = file_sep })

-- Controls length of printed strings when Debugging. 
M.max_print = 32000

function M.set_url(in_url)
   return rest_common.set_url(in_url),synch_rest_settings()
end

function M.set_auth(auth_method,user_name,pass_word)
   return rest_common.set_auth(auth_method,user_name,pass_word),synch_rest_settings()
end

function M.set_tgt(tgt)
   return rest_common.set_tgt(tgt),synch_rest_settings()
end

function M.get_tgt_cas( username, password )
   return rest_common.get_tgt_cas( username, password ),synch_rest_settings()
end

function M.get_service_ticket( request )
   return rest_common.get_service_ticket( request ),synch_rest_settings()
end
 
--- In the original rest_service, you could pass the body and header in as a fileref - preserve this funcitonality.
function M.request( action, request, body, header_in, content_type, out, hout )
   if type(body) == 'table' then
      body = utils.read_fileref(body[1])
      if body:len() == 0 then body = nil end
   end
   
   if type(header_in) == 'table' then
      header_in = utils.read_fileref(header_in[1])
      if header_in:len() == 0 then header_in = nil end
   end
   local pass, return_body, code = rest_common.request( action, request, body, header_in, content_type, out, hout )
   synch_rest_settings()
   return pass, return_body, code
end

--Alias the legacy tokenauth request function 
M.request_json_tokenauth = M.request

--Alias the legacy request2 function 
M.request2 = M.request

-- @Section HHTP Return Codes Table
M.http_codes={}
M.http_codes[100]='Continue'
M.http_codes[101]='Switching Protocols'
M.http_codes[102]='Processing (WebDAV; RFC 2518)'
M.http_codes[200]='OK'
M.http_codes[201]='Created'
M.http_codes[202]='Accepted'
M.http_codes[203]='Non-Authoritative Information (since HTTP/1.1)'
M.http_codes[204]='No Content'
M.http_codes[205]='Reset Content'
M.http_codes[206]='Partial Content (RFC 7233)'
M.http_codes[207]='Multi-Status (WebDAV; RFC 4918)'
M.http_codes[208]='Already Reported (WebDAV; RFC 5842)'
M.http_codes[226]='IM Used (RFC 3229)'
M.http_codes[300]='Multiple Choices'
M.http_codes[301]='Moved Permanently'
M.http_codes[302]='Found'
M.http_codes[303]='See Other (since HTTP/1.1)'
M.http_codes[304]='Not Modified (RFC 7232)'
M.http_codes[305]='Use Proxy (since HTTP/1.1)'
M.http_codes[306]='Switch Proxy'
M.http_codes[307]='Temporary Redirect (since HTTP/1.1)'
M.http_codes[308]='Permanent Redirect (RFC 7538)'
M.http_codes[400]='Bad Request'
M.http_codes[401]='Unauthorized (RFC 7235)'
M.http_codes[402]='Payment Required'
M.http_codes[403]='Forbidden'
M.http_codes[404]='Not Found'
M.http_codes[405]='Method Not Allowed'
M.http_codes[406]='Not Acceptable'
M.http_codes[407]='Proxy Authentication Required (RFC 7235)'
M.http_codes[408]='Request Timeout'
M.http_codes[409]='Conflict'
M.http_codes[410]='Gone'
M.http_codes[411]='Length Required'
M.http_codes[412]='Precondition Failed (RFC 7232)'
M.http_codes[413]='Payload Too Large (RFC 7231)'
M.http_codes[414]='URI Too Long (RFC 7231)'
M.http_codes[415]='Unsupported Media Type'
M.http_codes[416]='Range Not Satisfiable (RFC 7233)'
M.http_codes[417]='Expectation Failed'
M.http_codes[418]="I'm a teapot (RFC 2324)"
M.http_codes[421]='Misdirected Request (RFC 7540)'
M.http_codes[422]='Unprocessable Entity (WebDAV; RFC 4918)'
M.http_codes[423]='Locked (WebDAV; RFC 4918)'
M.http_codes[424]='Failed Dependency (WebDAV; RFC 4918)'
M.http_codes[426]='Upgrade Required'
M.http_codes[428]='Precondition Required (RFC 6585)'
M.http_codes[429]='Too Many Requests (RFC 6585)'
M.http_codes[431]='Request Header Fields Too Large (RFC 6585)'
M.http_codes[451]='Unavailable For Legal Reasons'
M.http_codes[500]='Internal Server Error'
M.http_codes[501]='Not Implemented'
M.http_codes[502]='Bad Gateway'
M.http_codes[503]='Service Unavailable'
M.http_codes[504]='Gateway Timeout'
M.http_codes[505]='HTTP Version Not Supported'
M.http_codes[506]='Variant Also Negotiates (RFC 2295)'
M.http_codes[507]='Insufficient Storage (WebDAV; RFC 4918)'
M.http_codes[508]='Loop Detected (WebDAV; RFC 5842)'
M.http_codes[510]='Not Extended (RFC 2774)'
M.http_codes[511]='Network Authentication Required (RFC 6585)'
M.http_codes[103]='Checkpoint'
M.http_codes[420]='Method Failure (Spring Framework)'
M.http_codes[450]='Blocked by Windows Parental Controls (Microsoft)'
M.http_codes[499]='Token Required or Request has been forbidden by antivirus]'
M.http_codes[509]='Bandwidth Limit Exceeded (Apache Web Server/cPanel)'
M.http_codes[530]='Site is frozen'
M.http_codes[440]='Login Timeout'
M.http_codes[449]='Retry With'
M.http_codes[451]='Redirect'
M.http_codes[444]='No Response'
M.http_codes[495]='SSL Certificate Error'
M.http_codes[496]='SSL Certificate Required'
M.http_codes[497]='HTTP Request Sent to HTTPS Port'
M.http_codes[499]='Client Closed Request'
M.http_codes[520]='Unknown Error'
M.http_codes[521]='Web Server Is Down'
M.http_codes[522]='Connection Timed Out'
M.http_codes[523]='Origin Is Unreachable'
M.http_codes[524]='A Timeout Occurred'
M.http_codes[525]='SSL Handshake Failed'
M.http_codes[526]='Invalid SSL Certificate'

return M
