/*--------------------------------------------------------------------------------
   Copyright (C), 2016 SAS Institute Inc.  All Rights Reserved
   sst_export_sheet_data:  Export sheet data from the Risk and Finance Workbench

   inputs:  scheme     - http or https, defaults to http.  See https://en.wikipedia.org/wiki/Uniform_Resource_Locator 
                                        for the syntax (names of parts) of a URL
            hostname   - hostname of the server running the RFW, defaults to localhost
            port       - port number of the server running the RFW, defaults to 7980
            project_id      - The project's ID (number) 
            sheet_name      - The sheet name for which the data should be exported
            output_filename - The output filename to place the exported data file into.  The output file is placed in the
                              directory determined by the RFW product's application setting named "exported.data.dir" points to.   
            iteration_number- the project's iteration to pull the data from.  If not specified, pulls from current iteration.  
            filter_likeys   - Any filtering to do on the data.  This variable is a concatenated parameter
                              of the form:  filterLiKey=<likey1>&filterLiKey=<likey2>&...
                              where <likeyN> are the likeys to filter data by.  Specify entity using this.  
            shock_name      - The name of a variable shock to use as additional filtering on the filter axis (optional)
            scenario_name   - The name of the scenario to use as additional filtering on the filter axis (optional)
            export_value_types - a concatenation of any of the following letters:
                                 N (the default) to only export numeric and resolved formula numeric values
                                 T to export text data
            userid     - the userid to use for the HTTP call into the RFW mid-tier to perform the operation.
                         The userid/password combination is optional when this macro is executed within a SAS session created by
                         the mid-tier for executing an RFW Model Group or Script.
            password   - the password to use for the HTTP call into the RFW mid-tier to perform the operation.
                         The userid/password combination is optional when this macro is executed within a SAS session created by
                         the mid-tier for executing an RFW Model Group or Script.
            outvar_http_status_code - an optional *global* macro variable to place the HTTP response's status code from the HTTP call to the RFW mid-tier 
            outvar_http_body - an optional *global* macro variable to place the HTTP response body from the HTTP call to the RFW mid-tier
            
            
   Outputs:   - The output excel file is located in the exported_data directory on the server tier.
   
   Notes:
              - This macro is meant to be used with a workspace server running SAS Host (basic) authentication.
                One can supply the userid and password via parameters to this macro or, if run within a Script or Model Group,
                the userid and password will be picked up automatically from the mid-tier execution.
              - The output excel spreadsheet can be read by sas by using the excel libname engine,
                which on Windows is called "xlsx" but on unix is called "sasioxlx".
              - Sometimes the excel libname engine can add $ signs to the sheet names making them
                tricky to use.  If that happens, here is an example of accessing the data from sas.
                Replace the file path with your file's path.
                
                libname exlib excel '/install/Config/Lev1/AppData/SASRiskandFinanceWorkbench/exported_data/Outputfile.xlsx' access=readonly;
                data your_data_set;
                  set exlib.'ExportedData$'n;
                run; 
------------------------------------------------------------------------------------*/

options NOQUOTELENMAX;
%macro sst_export_sheet_data(scheme=http, hostname=localhost, port=7980, project_id=, sheet_name=, output_filename=, iteration_number=, filter_likeys=, shock_name=, scenario_name=, export_value_types=N, userid=, password=, outvar_http_status_code=, outvar_http_body=);

  /* url encode the request parameters.  the urlencode function does not like empty strings */
  %if "&sheet_name" ne "" %then %do;
    %let sheet_name=%sysfunc(urlencode(&sheet_name));
  %end;
  %if "&output_filename" ne "" %then %do;
    %let output_filename=%sysfunc(urlencode(&output_filename));
  %end;
  %if "&scenario_name" ne "" %then %do;
    %let scenario_name=%sysfunc(urlencode(&scenario_name));
  %end;
  %if "&shock_name" ne "" %then %do;
    %let shock_name=%sysfunc(urlencode(&shock_name));
  %end;
  %if "&export_value_types" ne "" %then %do;
    %let export_value_types=%sysfunc(urlencode(&export_value_types));
  %end;
  

  %local iter_parm;
  %if "&iteration_number" ne "" %then %do;
      %let iter_parm=%nrstr(&)iteration%nrstr(=)&iteration_number;
  %end;
  %else
  %do;
      %let iter_parm=;
  %end;
  
  %local filter_parm;
  %if "&filter_likeys" ne "" %then %do;
      %let filter_parm=%nrstr(&)&filter_likeys;
  %end;
  %else
  %do; 
      %let filter_parm=;
  %end;
  
  %local scenario_parm;
  %if "&scenario_name" ne "" %then %do;
      %let scenario_parm=%nrstr(&)scenarioName%nrstr(=)&scenario_name;
  %end;
  %else
  %do; 
      %let scenario_parm=;
  %end;
  
  %local shock_parm;
  %if "&shock_name" ne "" %then %do;
      %let shock_parm=%nrstr(&)shockName%nrstr(=)&shock_name;
  %end;
  %else
  %do; 
      %let shock_parm=;
  %end;
  
  %local export_value_types_parm;
  %let export_value_types_parm=%nrstr(&)exportValueTypes%nrstr(=)%superq(export_value_types);
  
  
  %let url=%superq(scheme)://%superq(hostname):&port./SASRiskandFinanceWorkbench/rest/;
  %let request=projects/&project_id./sheets/byName?sheetName=%superq(sheet_name)%nrstr(&)outputFilename=%superq(output_filename)%superq(iter_parm)%superq(filter_parm)%superq(scenario_parm)%superq(shock_parm)%superq(export_value_types_parm);
  %let source=sst_export_sheet_data;

  proc lua restart infile="risk.stress.macro_api_entry"; run;
  
%mend;
